
/*
	exec_stream.h                              by Viktor "AlgorithMan" Engelmann
	                                             Viktor.Engelmann@RWTH-Aachen.de
	                                                   http://www.AlgorithMan.de
	thanks to
		http://www.pronix.de/pronix-202.html
		http://www.angelikalanger.com/Articles/C++Report/IOStreamsDerivation/IOStreamsDerivation.html
		http://www.angelikalanger.com/IOStreams/Excerpt/excerpt.htm
*/


#ifndef __EXECSTREAM_H
	#define __EXECSTREAM_H

	#ifndef EXECSTREAM_BUFFERSIZE
		#define EXECSTREAM_BUFFERSIZE 4096
	#endif

	#include <unistd.h>
	#include <sys/wait.h>
	#include <stdio.h>
	#include <fcntl.h>
	#include <stdlib.h>
	#include <limits.h>

	#include <sys/types.h>
	#include <signal.h>
	#include <iostream>
	using namespace std;



	class exec_streambuf : public streambuf {
		protected:
			int ParentReadChildWrite[2];
			int ParentWriteChildRead[2];
			int Status;
			pid_t Process;
			FILE* PipeWriter;
			FILE* PipeReader;
		public:
			exec_streambuf(string Command);
			~exec_streambuf();

			bool Alive();
			bool Pause();
			bool Continue();
			bool Close(int timeout_sec, int timeout_msec);
			bool Kill();
			void Wait();

			// overwritten virtual streambuf methods (to connect to streams)
			streamsize showmanyc();
			int_type overflow(int_type c);
			streamsize xsputn(const char* s, streamsize n);
			int_type underflow();
			streamsize xsgetn(char* s, streamsize n);
	};



	exec_streambuf::exec_streambuf(string Command) {
		// create pipes
		if(pipe(ParentReadChildWrite) == -1 || pipe(ParentWriteChildRead) == -1) {
			cerr << "Error creating pipe for call to external program \""<< Command << "\"\n";
			return;
		}
		// fork the process
		Process = fork();
		if(Process < 0) {
			cerr << "Error creating child-process for call to external program \""<<Command<<"\"\n";
			return;
		}
		if(Process > 0) { // Parent process
			close(ParentReadChildWrite[1]);
			close(ParentWriteChildRead[0]);
			PipeReader = fdopen(ParentReadChildWrite[0],"r");
			PipeWriter = fdopen(ParentWriteChildRead[1],"w");
			if(PipeWriter == NULL || PipeReader == NULL) {
				cerr << "Error establishing pipe to external program \""<<Command<<"\"\n";
				return;
			}

			// handling stuff derived from streambuf
			_M_in_beg = new char_type[EXECSTREAM_BUFFERSIZE+1];
			_M_in_cur = _M_in_beg;
			_M_in_end = _M_in_beg;
			_M_out_beg = new char_type[EXECSTREAM_BUFFERSIZE+1];
			_M_out_cur = _M_out_beg;
			_M_out_end = _M_out_beg + EXECSTREAM_BUFFERSIZE;
		} else { // Child process
			close(ParentReadChildWrite[0]);
			close(ParentWriteChildRead[1]);

			// set pipes to stdin and stdout of the external program
			if(ParentWriteChildRead[0] != STDIN_FILENO) {
				if(dup2(ParentWriteChildRead[0], STDIN_FILENO) != STDIN_FILENO) {
					cerr << "Error linking pipe to external program \""<<Command<<"\"s stdin\n";
					return;
				}
				close(ParentWriteChildRead[0]);
			}
			if(ParentReadChildWrite[1] != STDOUT_FILENO) {
				if(dup2(ParentReadChildWrite[1], STDOUT_FILENO) != STDOUT_FILENO) {
					cerr << "Error linking pipe to external program \""<<Command<<"\"s stdout\n";
					return;
				}
				close(ParentReadChildWrite[1]);
			}
			// start external program
			int result = system(Command.c_str());
			close(ParentReadChildWrite[1]);
			close(ParentWriteChildRead[0]);
		}
	}

	exec_streambuf::~exec_streambuf() {
		Close(2,0);
		delete[] _M_in_beg;
		delete[] _M_out_beg;
	}



	bool exec_streambuf::Alive() {
		int res = waitpid(Process, &Status, WNOHANG|WUNTRACED);
		if(res < 0)
			return false;
		if(res == 0)
			return true;
		return !WIFEXITED(Status) && !WIFSIGNALED(Status);
	}

	bool exec_streambuf::Close(int timeout_sec, int timeout_msec) {
		if(!Alive())
			return false;
		kill(Process, SIGTERM);
		fclose(PipeWriter);
		fclose(PipeReader);
		close(ParentReadChildWrite[0]);
		close(ParentWriteChildRead[1]);
		// alarm(2); ???
		if (waitpid(Process, NULL, WNOHANG) == -1) {
			cerr << "Error: Child Process didn\'t terminate on time - trying to kill it"; // ---- not working yet
			return Kill();
		}
		return true;
	}

	bool exec_streambuf::Kill() {
		if(!Alive())
			return false;
		return kill(Process, SIGKILL) == 0;
	}

	bool exec_streambuf::Pause() {
		if(!Alive())
			return false;
		return kill(Process, SIGSTOP) == 0;
	}

	bool exec_streambuf::Continue() {
		if(!Alive())
			return false;
		return kill(Process, SIGCONT) == 0;
	}

	void exec_streambuf::Wait() {
		waitpid(Process,&Status,0);
	}



	streamsize exec_streambuf::showmanyc() {
		return _M_in_end - _M_in_cur;
	}

	streamsize exec_streambuf::xsgetn(char* s, streamsize n) {
		if(_M_in_cur + n <= _M_in_end) { // request can be satisfied
			memcpy(s, _M_in_cur, sizeof(char_type)*n);
			_M_in_cur += n;
			return n;
		}

		int avail = showmanyc(); // copy remaining 
		if(avail > 0) {
			memcpy(s, _M_in_cur, sizeof(char_type)*avail);
			_M_in_cur = _M_in_end;
		}
		if(underflow() != traits_type::eof()) // copy from refilled buffer
			return avail+xsgetn(s+avail, n-avail);
		else
			return avail;
	}

	exec_streambuf::int_type exec_streambuf::underflow() {
		int read = fread(_M_in_beg, sizeof(char_type), EXECSTREAM_BUFFERSIZE, PipeReader); // is this blocking??
		if(read <= 0) {
			_M_in_end = _M_in_beg;
			_M_in_cur = _M_in_beg;
			return traits_type::eof();
		}
		_M_in_end = _M_in_beg + read;
		_M_in_cur = _M_in_beg;
		return *_M_in_beg;
	}

	streamsize exec_streambuf::xsputn(const char* s, streamsize n) {
		if(_M_out_cur+n <= _M_out_end) { // request can be satisfied
			memcpy(_M_out_cur, s, sizeof(char_type)*n);
			_M_out_cur += n+1;
			return 1;
		}
		int availspace = _M_out_end - _M_out_cur;
		if(availspace > 0) {
			memcpy(_M_out_cur, s, sizeof(char_type)*availspace);
			_M_out_cur = _M_out_end;
		}
		if(overflow(s[availspace]) != traits_type::eof()) {
			return xsputn(s+availspace, n-availspace);
		} else
			return traits_type::eof();
	}


	exec_streambuf::int_type exec_streambuf::overflow(int_type c) {
		*_M_out_cur = '\0'; // flush buffer
		if(fwrite(_M_out_beg, sizeof(char_type), 1+_M_out_cur-_M_out_beg, PipeWriter) > 0) {
			fflush(PipeWriter);

			_M_out_cur = _M_out_beg+1; // reset buffer
			*_M_out_beg = c;
			return 1;
		} else {
			return traits_type::eof();
		}
	}





	class execstream : public iostream {
		private:
			exec_streambuf msb;
		public:
			execstream(string command) :
				msb(command),
				iostream(&msb),
				ios(0) {
			}
	};

#endif
